import logging

from twisted.internet.defer import inlineCallbacks, returnValue
import yaml
import zookeeper

from juju.errors import JujuError, ProviderError
from juju.providers.common.utils import get_active_machine_provider
from juju.state.errors import UnitRelationStateNotFound
from juju.state.charm import CharmStateManager
from juju.state.machine import MachineStateManager
from juju.state.relation import RelationStateManager
from juju.state.service import ServiceStateManager
from juju.state.base import StateBase
from juju.unit.workflow import WorkflowStateClient

# Version of the status serialisation
VERSION = 1

log = logging.getLogger("juju.state.status")


class StatusManager(StateBase):
    """Watch deployments for status state changes."""
    def __init__(self, client, callback=None):
        super(StatusManager, self).__init__(client)
        self._status = None
        self._callback = callback

    @inlineCallbacks
    def get_status(self):
        """Get current status information."""
        if self._status is None:
            status = yield self._load_status()
            if status is None:
                status = yield self.collect()

            self._status = status
        returnValue(self._status)

    @inlineCallbacks
    def collect(self):
        """Collect and cache all status state.

        This method typically is called on topology change and creates
        a version of the status which is held in this manager and returned
        until the next change.
        """
        charm_manager = CharmStateManager(self._client)
        machine_manager = MachineStateManager(self._client)
        relation_manager = RelationStateManager(self._client)
        service_manager = ServiceStateManager(self._client)

        machine_provider = yield get_active_machine_provider(self._client)

        service_data = {}
        machine_data = {}

        state = dict(services=service_data, machines=machine_data)
        services = yield service_manager.get_all_service_states()
        for service in services:
            unit_data = {}
            relation_data = {}
            charm_id = yield service.get_charm_id()
            charm = yield charm_manager.get_charm_state(charm_id)

            service_data[service.service_name] = dict(units=unit_data,
                                                      charm=charm.id,
                                                      relations=relation_data)
            exposed = yield service.get_exposed_flag()
            if exposed:
                service_data[service.service_name].update(exposed=exposed)

            units = yield service.get_all_unit_states()
            relations = yield relation_manager.get_relations_for_service(service)

            for unit in units:
                u = unit_data[unit.unit_name] = dict()
                u["public-address"] = yield unit.get_public_address()
                machine_id = yield unit.get_assigned_machine_id()
                u["machine"] = machine_id

                unit_workflow_client = WorkflowStateClient(self._client, unit)
                u["state"] = yield unit_workflow_client.get_state()
                if exposed:
                    open_ports = yield unit.get_open_ports()
                    u["open-ports"] = ["{port}/{proto}".format(**port_info)
                                       for port_info in open_ports]

                # collect info on each relation for the service unit
                relation_status = {}
                for relation in relations:
                    try:
                        relation_unit = yield relation.get_unit_state(unit)
                    except UnitRelationStateNotFound:
                        # This exception will occur when relations are
                        # established between services without service
                        # units, and therefore never have any
                        # corresponding service relation units. This
                        # scenario does not occur in actual deployments,
                        # but can happen in test circumstances. In
                        # particular, it will happen with a misconfigured
                        # provider, which exercises this codepath.
                        continue  # should not occur, but status should not fail
                    relation_workflow_client = WorkflowStateClient(
                        self._client, relation_unit)
                    relation_workflow_state = \
                        yield relation_workflow_client.get_state()
                    relation_status[relation.relation_name] = dict(
                        state=relation_workflow_state)
                u["relations"] = relation_status

            for relation in relations:
                rel_services = yield relation.get_service_states()

                # A single related service implies a peer relation. More
                # imply a bi-directional provides/requires relationship.
                # In the later case we omit the local side of the relation
                # when reporting.
                if len(rel_services) > 1:
                    # Filter out self from multi-service relations.
                    rel_services = [
                        rsn for rsn in rel_services if rsn.service_name !=
                        service.service_name]

                if len(rel_services) > 1:
                    raise ValueError("Unexpected relationship with more "
                                     "than 2 endpoints")

                rel_service = rel_services[0]
                relation_data[relation.relation_name] = rel_service.service_name

        machines = yield machine_manager.get_all_machine_states()
        for machine_state in machines:
            instance_id = yield machine_state.get_instance_id()
            m = {"instance-id": instance_id if instance_id is not None else "pending"}
            if instance_id is not None:
                try:
                    pm = yield machine_provider.get_machine(instance_id)
                    m["dns-name"] = pm.dns_name
                except ProviderError:
                    # The provider doesn't have machine information
                    log.error("Machine provider information missing: machine %s" % (
                              machine_state.id))

            machine_data[machine_state.id] = m

        yield self._store_status(state)
        returnValue(state)

    @inlineCallbacks
    def _load_status(self):
        try:
            ystatus, _ = yield self._client.get("/status")
        except zookeeper.NoNodeException:
            returnValue(None)

        status = yaml.load(ystatus)
        if status is not None and "version" not in status or status["version"] != VERSION:
            raise JujuError("Version mismatch in status")
        del status["version"]

        returnValue(status)

    @inlineCallbacks
    def _store_status(self, state):
        versioned_state = dict(state)
        versioned_state["version"] = VERSION
        ys = yaml.dump(versioned_state)

        exists = yield self._client.exists("/status")
        if not exists:
            yield self._client.create("/status", ys)
        else:
            yield self._client.set("/status", ys)

        self._status = state

    def watch_status(self):
        """Set a permanent watch to refresh status on change."""

        @inlineCallbacks
        def watch_topology(old_topology, new_topology):
            state = yield self.collect()
            if self._callback:
                self._callback(state)
            returnValue(state)

        return self._watch_topology(watch_topology)
