from twisted.internet.defer import inlineCallbacks, returnValue
import yaml

from juju.errors import JujuError
from juju.environment.environment import Environment
from juju.environment.tests.test_config import EnvironmentsConfigTestBase
from juju.state.endpoint import RelationEndpoint
from juju.state.environment import EnvironmentStateManager
from juju.state.status import StatusManager
from juju.state.tests.test_service import ServiceStateManagerTestBase
from juju.unit.workflow import ZookeeperWorkflowState


class StatusTestBase(ServiceStateManagerTestBase, EnvironmentsConfigTestBase):

    @inlineCallbacks
    def setUp(self):
        yield super(StatusTestBase, self).setUp()
        config = {
            "juju": "environments",
            "environments": {
                "firstenv": {
                    "type": "dummy",
                    "admin-secret": "homer"}}}
        self.write_config(yaml.dump(config))
        self.config.load()

        self.environment = self.config.get_default()
        env_manager = EnvironmentStateManager(self.client)
        yield env_manager.set_config_state(self.config, "firstenv")
        self.provider = yield self.environment.get_machine_provider()

    def mock_environment(self):
        mock_environment = self.mocker.patch(Environment)
        mock_environment.get_machine_provider()
        self.mocker.result(self.provider)
        self.mocker.count(min=0, max=None)

    @inlineCallbacks
    def set_unit_state(self, unit_state, state, port_protos=()):
        yield unit_state.set_public_address(
            "%s.example.com" % unit_state.unit_name.replace("/", "-"))
        workflow_client = ZookeeperWorkflowState(self.client, unit_state)
        yield workflow_client.set_state(state)
        for port_proto in port_protos:
            yield unit_state.open_port(*port_proto)

    @inlineCallbacks
    def add_relation_unit_states(self, relation_state, unit_states, states):
        for unit_state, state in zip(unit_states, states):
            relation_unit_state = yield relation_state.add_unit_state(
                unit_state)
            workflow_client = ZookeeperWorkflowState(
                self.client, relation_unit_state)
            yield workflow_client.set_state(state)

    @inlineCallbacks
    def add_relation_with_relation_units(
        self,
        source_endpoint, source_units, source_states,
        dest_endpoint, dest_units, dest_states):
        relation_state, service_relation_states = \
            yield self.relation_state_manager.add_relation_state(
            *[source_endpoint, dest_endpoint])
        source_relation_state, dest_relation_state = service_relation_states
        yield self.add_relation_unit_states(
            source_relation_state, source_units, source_states)
        yield self.add_relation_unit_states(
            dest_relation_state, dest_units, dest_states)

    @inlineCallbacks
    def set_unit_state(self, unit_state, state, port_protos=()):
        unit_state.set_public_address(
            "%s.example.com" % unit_state.unit_name.replace("/", "-"))
        workflow_client = ZookeeperWorkflowState(self.client, unit_state)
        yield workflow_client.set_state(state)
        for port_proto in port_protos:
            yield unit_state.open_port(*port_proto)

    @inlineCallbacks
    def build_topology(self, base=None):
        """Build a simulated topology with a default machine configuration.

        This method returns a dict that can be used to get handles to
        the constructed objects.
        """
        state = {}

        # build out the topology using the state managers

        # inform the provider about the machine
        state["machines"] = machine_list = []
        for i in range(7):
            m = yield self.machine_state_manager.add_machine_state()
            machine_list.append(m)
            public_address = "steamcloud-%d.com" % (i + 1)
            yield self.provider.start_machine({"machine-id": i,
                                               "dns-name": public_address})

            yield m.set_instance_id(i)

        # "Deploy" services
        wordpress = yield self.add_service_from_charm("wordpress")
        mysql = yield self.add_service_from_charm("mysql")
        yield mysql.set_exposed_flag()  # but w/ no open ports

        varnish = yield self.add_service_from_charm("varnish")
        yield varnish.set_exposed_flag()
        # w/o additional metadata
        memcache = yield self.add_service("memcache")

        state["services"] = dict(wordpress=wordpress, mysql=mysql,
                                 varnish=varnish, memcache=memcache)

        # add unit states to services and assign to machines
        wpu = yield wordpress.add_unit_state()
        yield wpu.assign_to_machine(machine_list[0])

        myu1 = yield mysql.add_unit_state()
        myu2 = yield mysql.add_unit_state()
        yield myu1.assign_to_machine(machine_list[1])
        yield myu2.assign_to_machine(machine_list[2])

        vu1 = yield varnish.add_unit_state()
        vu2 = yield varnish.add_unit_state()
        yield vu1.assign_to_machine(machine_list[3])
        yield vu2.assign_to_machine(machine_list[4])

        mc1 = yield memcache.add_unit_state()
        mc2 = yield memcache.add_unit_state()
        yield mc1.assign_to_machine(machine_list[5])
        yield mc2.assign_to_machine(machine_list[6])

        # Set the lifecycle state and open ports, if any, for each unit state.
        yield self.set_unit_state(wpu, "started", [(80, "tcp"), (443, "tcp")])
        yield self.set_unit_state(myu1, "started")
        yield self.set_unit_state(myu2, "stopped")
        yield self.set_unit_state(vu1, "started", [(80, "tcp")])
        yield self.set_unit_state(vu2, "started", [(80, "tcp")])
        yield self.set_unit_state(mc1, "started")
        yield self.set_unit_state(mc2, "installed")

        # Wordpress integrates with each of the following
        # services. Each relation endpoint is used to define the
        # specific relation to be established.
        mysql_ep = RelationEndpoint(
            "mysql", "client-server", "db", "server")
        memcache_ep = RelationEndpoint(
            "memcache", "client-server", "cache", "server")
        varnish_ep = RelationEndpoint(
            "varnish", "client-server", "proxy", "client")

        wordpress_db_ep = RelationEndpoint(
            "wordpress", "client-server", "db", "client")
        wordpress_cache_ep = RelationEndpoint(
            "wordpress", "client-server", "cache", "client")
        wordpress_proxy_ep = RelationEndpoint(
            "wordpress", "client-server", "proxy", "server")

        # Create relation service units for each of these relations
        yield self.add_relation_with_relation_units(
            mysql_ep, [myu1, myu2], ["up", "departed"],
            wordpress_db_ep, [wpu], ["up"])
        yield self.add_relation_with_relation_units(
            memcache_ep, [mc1, mc2], ["up", "down"],
            wordpress_cache_ep, [wpu], ["up"])
        yield self.add_relation_with_relation_units(
            varnish_ep, [vu1, vu2], ["up", "up"],
            wordpress_proxy_ep, [wpu], ["up"])

        state["relations"] = dict(
            wordpress=[wpu],
            mysql=[myu1, myu2],
            varnish=[vu1, vu2],
            memcache=[mc1, mc2]
            )
        returnValue(state)

    def validate_topology(self, state):
        services = state["services"]
        self.assertIn("wordpress", services)
        self.assertIn("varnish", services)
        self.assertIn("mysql", services)

        # and verify the specifics of a single service
        self.assertTrue("mysql" in services)
        units = list(services["mysql"]["units"])
        self.assertEqual(len(units), 2)

        self.assertEqual(state["machines"][0],
                         {"instance-id": 0,
                          "dns-name": "steamcloud-1.com"})

        self.assertEqual(services["mysql"]["relations"],
                         {"db": "wordpress"})

        self.assertEqual(services["wordpress"]["relations"],
                         {"cache": "memcache",
                          "db": "mysql",
                          "proxy": "varnish"})

        self.assertEqual(
            services["varnish"],
            {"units":
                 {"varnish/1": {
                        "machine": 4,
                        "state": "started",
                        "public-address": "varnish-1.example.com",
                        "open-ports": ["80/tcp"],
                        "relations": {"proxy": {"state": "up"}}},
                  "varnish/0": {
                        "machine": 3,
                        "state": "started",
                        "public-address": "varnish-0.example.com",
                        "open-ports": ["80/tcp"],
                        "relations": {"proxy": {"state": "up"}}}},
             "exposed": True,
             "charm": "local:series/varnish-1",
             "relations": {"proxy": "wordpress"}})

        self.assertEqual(
            services["wordpress"],
            {"charm": "local:series/wordpress-3",
             "relations": {
                    "cache": "memcache",
                    "db": "mysql",
                    "proxy": "varnish"},
             "units": {
                    "wordpress/0": {
                        "machine": 0,
                        "public-address": "wordpress-0.example.com",
                        "relations": {
                            "cache": {"state": "up"},
                            "db": {"state": "up"},
                            "proxy": {"state": "up"}},
                        "state": "started"}}})

    @inlineCallbacks
    def test_collect(self):
        yield self.build_topology()

        # we need the provider that knows about our dummy machines
        self.mock_environment()
        self.mocker.replay()

        # collect everything
        sm = StatusManager(self.client)
        state = yield sm.collect()
        self.validate_topology(state)

    @inlineCallbacks
    def test_get_status(self):
        yield self.build_topology()

        # we need the provider that knows about our dummy machines
        self.mock_environment()
        self.mocker.replay()

        # collect everything
        sm = StatusManager(self.client)
        state = yield sm.get_status()
        self.validate_topology(state)

    @inlineCallbacks
    def test_watch_status(self):
        yield self.build_topology()

        # we need the provider that knows about our dummy machines
        self.mock_environment()
        self.mocker.replay()

        # collect everything
        sm = StatusManager(self.client)

        yield sm.watch_status()

        state = yield sm.get_status()
        self.validate_topology(state)

        # now make a change, and see if the manager has it
        wp_service = yield self.service_state_manager.get_service_state("wordpress")
        self.assertEqual(len(state["services"]["wordpress"]["units"]), 1)

        # create a new unit
        yield wp_service.add_unit_state()
        yield self.sleep(1)
        # validate we see the change
        state2 = yield sm.get_status()
        self.assertTrue(state2 is not state)

        wordpress = state2["services"]["wordpress"]
        self.assertEqual(len(wordpress["units"]), 2)

        # and a final unit
        yield wp_service.add_unit_state()
        yield self.sleep(1)
        state3 = yield sm.get_status()
        wordpress = state3["services"]["wordpress"]
        self.assertEqual(len(wordpress["units"]), 3)

    @inlineCallbacks
    def test_watch_status_with_callback(self):
        yield self.build_topology()

        # we need the provider that knows about our dummy machines
        self.mock_environment()
        self.mocker.replay()

        # a callback for status changes
        status_state = {}

        def status_change(state):
            status_state.clear()
            status_state.update(state)

        # collect everything
        sm = StatusManager(self.client, status_change)

        yield sm.watch_status()

        state = yield sm.get_status()
        self.validate_topology(state)
        # validate that the callback set our local to the proper state
        self.assertEqual(status_state, state)

        # now make a change, and see if the manager has it
        wp_service = yield self.service_state_manager.get_service_state("wordpress")
        self.assertEqual(len(state["services"]["wordpress"]["units"]), 1)

        # create a new unit
        yield wp_service.add_unit_state()
        yield self.sleep(1)
        # validate we see the change
        state2 = yield sm.get_status()
        self.assertTrue(state2 is not state)

        wordpress = state2["services"]["wordpress"]
        self.assertEqual(len(wordpress["units"]), 2)

        # and a final unit
        yield wp_service.add_unit_state()
        yield self.sleep(1)
        state3 = yield sm.get_status()
        wordpress = state3["services"]["wordpress"]
        self.assertEqual(len(wordpress["units"]), 3)

        self.assertEqual(status_state, state3)

    @inlineCallbacks
    def test_load_status(self):
        # even though this is a private method we test its logic
        # directly
        self.mock_environment()
        self.mocker.replay()

        sm = StatusManager(self.client)
        result = yield sm._load_status()
        self.assertEqual(result, None)

        # Store an empty dict
        yield sm._store_status({})
        result = yield sm._load_status()
        self.assertEqual(result, {})

        # now write the status node directly w/o version
        yield self.client.set("/status", yaml.dump({}))
        error = sm._load_status()
        error = yield self.assertFailure(error, JujuError)
        self.assertEqual(str(error), "Version mismatch in status")

        # likewise writing an invalid version produces an error
        yield self.client.set("/status", yaml.dump({"version": "ZZZ"}))
        error = sm._load_status()
        error = yield self.assertFailure(error, JujuError)
        self.assertEqual(str(error), "Version mismatch in status")

    @inlineCallbacks
    def test_store_status(self):
        # even though this is a private method we test its logic
        # directly
        from juju.state.status import VERSION
        self.mock_environment()
        self.mocker.replay()

        sm = StatusManager(self.client)
        yield sm._store_status({})

        # verify that we can find the encoded version inside the yaml
        # node
        data, _ = yield self.client.get("/status")
        state = yaml.load(data)
        self.assertEqual(state, {"version": VERSION})



